/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <algorithm>
#include <memory>

#include "common.hpp"


namespace /* anonymous */
{
  void
  example(const std::size_t n,
          const std::size_t k,
          const std::size_t m,
          const double *const restrict a_ptr,
          const double *const restrict b_ptr,
          double *restrict c_ptr,
          const double alpha,
          const double beta) noexcept
  {
    if (beta == 0.0)
      {
        std::uninitialized_fill(c_ptr, c_ptr + n * m, 0.0);
      }
    else
      {
#if HAVE_OPENMP && defined(_OPENMP)
#pragma omp parallel for
#endif
        for (std::size_t idx = 0; idx < n * m; ++idx)
          c_ptr[idx] *= beta;
      }
    for (std::size_t l = 0; l < k; ++l)
      {
#if HAVE_OPENMP && defined(_OPENMP)
#pragma omp parallel for
#endif
        for (std::size_t ij = 0; ij < n * m; ++ij)
          {
            const std::size_t i {ij % n};
            const std::size_t j {ij / n};
            c_ptr[i + n * j] += alpha * a_ptr[i + n * l] * b_ptr[j + m * l];
          }
      }
  }
}

namespace blas_benchmarks
{
  class Benchmark : public AbstractBenchmark
  {
  private:

    std::size_t n_ {};
    std::size_t k_ {};
    std::size_t m_ {};
    std::unique_ptr<double[]> aUptr_ {};
    std::unique_ptr<double[]> bUptr_ {};
    std::unique_ptr<double[]> cUptr_ {};

  protected:

    virtual void
    setUp(const std::size_t n,
          const std::size_t k,
          const std::size_t m) final override
    {
      this->n_ = n;
      this->k_ = k;
      this->m_ = m;
      this->aUptr_.reset(new double[n * k]);
      this->bUptr_.reset(new double[m * k]);
      this->cUptr_.reset(new double[n * m]);
      this->fillRandom_(this->aUptr_.get(), this->n_, this->k_);
      this->fillRandom_(this->bUptr_.get(), this->m_, this->k_);
      this->fillRandom_(this->cUptr_.get(), this->n_, this->m_);
    }

    virtual void
    compute(const double alpha, const double beta) final override
    {
      example(this->n_, this->k_, this->m_,
              this->aUptr_.get(), this->bUptr_.get(), this->cUptr_.get(),
              alpha, beta);
    }

    virtual void
    tearDown() final override
    {
      this->n_ = 0;
      this->k_ = 0;
      this->m_ = 0;
      this->aUptr_.reset();
      this->bUptr_.reset();
      this->cUptr_.reset();
    }

    virtual void
    selfTest(const std::size_t n,
             const std::size_t k,
             const std::size_t m,
             const test_matrix_type& a,
             const test_matrix_type& b,
             const test_matrix_type& c,
             const double alpha,
             const double beta,
             test_matrix_type& result) final override
    {
      this->setUp(n, k, m);
      for (std::size_t i = 0; i < n; ++i)
        for (std::size_t j = 0; j < k; ++j)
          this->aUptr_[i + j * n] = a[i][j];
      for (std::size_t i = 0; i < m; ++i)
        for (std::size_t j = 0; j < k; ++j)
          this->bUptr_[i + j * m] = b[i][j];
      for (std::size_t i = 0; i < n; ++i)
        for (std::size_t j = 0; j < m; ++j)
          this->cUptr_[i + j * n] = c[i][j];
      this->compute(alpha, beta);
      for (std::size_t i = 0; i < n; ++i)
        for (std::size_t j = 0; j < m; ++j)
          result[i][j] = this->cUptr_[i + j * n];
      this->tearDown();
    }

  private:

    void
    fillRandom_(double * m_ptr, const std::size_t n, const std::size_t m)
    {
      for (std::size_t j = 0; j < m; ++j)
        for (std::size_t i = 0; i < n; ++i)
          m_ptr[j * n + i] = this->random();
    }
  };
}


int
main(int argc, char * * argv)
{
  std::vector<std::string> args {};
  blas_benchmarks::Benchmark bm {};
  for (int i = 1; i < argc; ++i)
    args.push_back(argv[i]);
  return blas_benchmarks::common_main(args, bm);
}
