/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


#include <chrono>
#include <cmath>
#include <cstddef>
#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <limits>
#include <random>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>


namespace blas_benchmarks
{
  namespace common_hpp
  {
    template<typename T>
    T
    lexical_argument_cast(const std::string& arg,
                          const std::string& expected = "valid argument")
    {
      T value {};
      std::istringstream iss {arg};
      iss >> value;
      if (!iss || !iss.eof())
        {
          std::ostringstream oss {};
          oss << "not a " << expected << ": '" << arg << "'";
          throw std::invalid_argument {oss.str()};
        }
      return value;
    }

    class Random
    {
    private:

      std::default_random_engine engine_ {0};
      std::exponential_distribution<double> distMag_ {1.0};
      std::bernoulli_distribution distSig_ {};

    public:

      Random()
      {
        std::random_device rnd_dev {};
        this->engine_.seed(rnd_dev());
      }

      double
      operator()() noexcept
      {
        const double magnitude {distMag_(this->engine_)};
        return this->distSig_(this->engine_) ? magnitude : -magnitude;
      }
    };
  }

  class AbstractBenchmark
  {
  protected:

    using test_matrix_type = std::vector<std::vector<double>>;

  private:

    common_hpp::Random rnd_{};

  public:

    bool
    run(const unsigned samples,
        const std::size_t min_size,
        const std::size_t max_size)
    {
      const bool self_test_passed = this->runSelfTest_();
      AbstractBenchmark::printPreamble_(samples, min_size, max_size);
      std::default_random_engine rnd_engine {0};
      std::uniform_int_distribution<std::size_t> size_dist {min_size, max_size};
      {
        std::random_device rnd_dev {};
        rnd_engine.seed(rnd_dev());
      }
      for (unsigned s = 0; s < samples; ++s)
        {
          const std::size_t k {size_dist(rnd_engine)};
          const std::size_t n {size_dist(rnd_engine)};
          const std::size_t m {size_dist(rnd_engine)};
          this->setUp(n, k, m);
          const auto t0 = std::chrono::steady_clock::now();
          this->compute(this->random(), this->random());
          const auto t1 = std::chrono::steady_clock::now();
          this->tearDown();
          const auto seconds = std::chrono::duration_cast<std::chrono::duration<double>>(t1 - t0);
          std::cout << std::setw(16) << n
                    << std::setw(16) << k
                    << std::setw(16) << m
                    << std::setw(16) << std::setprecision(6) << std::scientific << seconds.count()
                    << std::endl;
        }
      return self_test_passed;
    }

  protected:

    virtual void
    setUp(const std::size_t n, const std::size_t k, const std::size_t m) = 0;

    virtual void
    compute(const double alpha, const double beta) = 0;

    virtual void
    tearDown() = 0;

    virtual void
    selfTest(const std::size_t n,
             const std::size_t k,
             const std::size_t m,
             const test_matrix_type& a,
             const test_matrix_type& b,
             const test_matrix_type& c,
             const double alpha,
             const double beta,
             test_matrix_type& result) = 0;

    inline double
    random() noexcept
    {
      return this->rnd_();
    }

  private:

    bool
    runSelfTest_()
    {
      constexpr double tol = 10.0 * std::numeric_limits<double>::epsilon();
      constexpr double nan = std::numeric_limits<double>::quiet_NaN();
      // Random test data generated and result computed with GNU Octave.
      const std::size_t n {7};
      const std::size_t k {5};
      const std::size_t m {6};
      const test_matrix_type a {
        {0.7898383959914768,  0.8336615368918433,  0.4332269019575177,  0.2093522933601948,  0.9025044556557740},
        {0.3462848753776698,  0.3291753193649838,  0.5823641020676487,  0.5389924862903379,  0.2076382650285426},
        {0.08899818946511488, 0.2311511424942966,  0.9593751318634506,  0.1176893758919291,  0.8596736553502752},
        {0.8327535618446061,  0.4454892868681083,  0.1232312812426644,  0.9951868769410481,  0.4697754205342465},
        {0.6602145748361358,  0.6313154173378537,  0.9325387002800224,  0.5038780344305265,  0.3111046160290944},
        {0.8182622476324708,  0.4424451884587522,  0.2997553127213941,  0.6736246113642538,  0.2798418958364997},
        {0.6299751005326488,  0.07073079748192328, 0.5441350861118592,  0.02979707519759327, 0.3470493206202921},
      };
      const test_matrix_type b {
        {0.8937749339075191,  0.968877174554588,   0.5198696580466965,  0.2916533618541232,  0.5396069667418728},
        {0.7345860545184723,  0.4990419237397284,  0.4058275237166193,  0.3869504786943795,  0.2436936831467733},
        {0.8798911919115209,  0.9437135114140484,  0.1630462767222743,  0.4022342216852582,  0.5230855272637905},
        {0.6758271466230755,  0.6120344296932202,  0.8425402263700063,  0.5766506394090462,  0.7472837390797954},
        {0.7836499147595960,  0.8311626539702458,  0.8968095951760812,  0.7455606805292804,  0.1255122950030852},
        {0.09158402138477344, 0.6439961502557613,  0.8016462395733481,  0.05592493153436677, 0.2288878997417598},
      };
      const test_matrix_type c {
        {0.5188001114948637,  0.5709492727321156,  0.2627213793802878,  0.4407022995793584,  0.8634363580252170,  0.1386401163916935},
        {0.08194743219526028, 0.6617990115122612,  0.8554693078413894,  0.5531261205933081,  0.6832552559810885,  0.6537120743624920},
        {0.2699702625751629,  0.4810564414986142,  0.3801208460326954,  0.9299492948987333,  0.8048711436022618,  0.5130689817106338},
        {0.2831643246414402,  0.6555603834731705,  0.2317316459779955,  0.3158429038865994,  0.3715670199487116,  0.6421045970059023},
        {0.8882890736466172,  0.9252442510150574,  0.05328451333944688, 0.9238573275755797,  0.9005843949884578,  0.9071190857914165},
        {0.9281569305740895,  0.9630923950653715,  0.3724849536782129,  0.8313734322861299,  0.4766074044420424,  0.8044286889134852},
        {0.7276237602175067,  0.3623081029478876,  0.3555999012801012,  0.5073922932769299,  0.03341390826813111, 0.8651075790707657},
      };
      const double alpha {0.01412540846073722};
      const double beta  {0.2985638899182497};
      const test_matrix_type expected {
        {0.1871988125736218,  0.1912714965657333,  0.1082245354252131,  0.1627128052155467,  0.2856144045298751,  0.05798727127338102},
        {0.04142306083507032, 0.2105020468889521,  0.2700420188066468,  0.1848079700411288,  0.2251149065713989,  0.2063087605178238},
        {0.09897292635190345, 0.1552810808939731,  0.1269079887840586,  0.3019480578128986,  0.2589211747057885,  0.1691376919775159},
        {0.1097385083993774,  0.2152709542847814,  0.09488463528906939, 0.1206320170624319,  0.1382594293259478,  0.2005394609183821},
        {0.2934813632187211,  0.2967161555241242,  0.03983947922633404, 0.3060775147662401,  0.3012734014053539,  0.2893935547297450},
        {0.3006092256954718,  0.3055175945045846,  0.133863945355906,   0.2718628726807429,  0.1679374975049742,  0.2500879163126162},
        {0.2329273922849943,  0.1196842859128383,  0.1189287745116159,  0.1684963825237859,  0.02560211982463154, 0.2670554295246971},
      };
      test_matrix_type actual {
        {nan, nan, nan, nan, nan, nan},
        {nan, nan, nan, nan, nan, nan},
        {nan, nan, nan, nan, nan, nan},
        {nan, nan, nan, nan, nan, nan},
        {nan, nan, nan, nan, nan, nan},
        {nan, nan, nan, nan, nan, nan},
        {nan, nan, nan, nan, nan, nan},
      };
      std::cerr << "Running self-test... ";
      std::size_t bad_count {0};
      this->selfTest(n, k, m, a, b, c, alpha, beta, actual);
      for (std::size_t i = 0; i < n; ++i)
        for (std::size_t j = 0; j < m; ++j)
          if (!std::isfinite(actual[i][j])
              || std::abs(expected[i][j] - actual[i][j]) >= tol)
            bad_count += 1;
      std::cerr << (!bad_count ? "passed" : "failed") << std::endl;
      return !bad_count;
    }

    static void
    printPreamble_(const unsigned samples,
                   const std::size_t min_size,
                   const std::size_t max_size)
    {
      std::vector<std::string> preamble_lines {
        "Execution time for computing",
        "",
        "    C := alpha * A * B^T + beta * C",
        "",
        "where A is a (n, k), B a (m, k) and C a (n, m) real matrix and alpha",
        "and beta are real scalars.  The dimensions n, m and k are individually",
        "chosen randomly for each sample from the range [min_size,, max_size].",
        "",
      };
      {
        std::ostringstream oss {};
        oss << "Parameters: "
            << "samples = " << samples << ", "
            << "min_size = " << min_size << ", "
            << "max_size = " << max_size << ".";
        preamble_lines.push_back(oss.str());
        preamble_lines.push_back("");
      }
      {
        std::ostringstream oss {};
        oss << std::setw(14) << "n / 1"
            << std::setw(16) << "k / 1"
            << std::setw(16) << "m / 1"
            << std::setw(16) << "time / s";
        preamble_lines.push_back(oss.str());
        preamble_lines.push_back("");
      }
      for (const auto& line : preamble_lines)
        std::cout << "# " << line << "\n";
      std::cout << std::flush;
    }
  };


  int
  common_main(const std::vector<std::string>& args, AbstractBenchmark& abm)
  {
    using common_hpp::lexical_argument_cast;
    std::size_t min_size { 100};
    std::size_t max_size {1000};
    unsigned samples     {   3};
    std::cin.exceptions(std::istream::failbit);
    std::cout.exceptions(std::ostream::failbit);
    std::cerr.exceptions(std::ostream::failbit);
    std::clog.exceptions(std::ostream::failbit);
    for (const auto arg : args)
      {
        if (arg == "--help")
          {
            std::cout << "usage: (0) " << PACKAGE << "\n"
                      << "       (1) " << PACKAGE << " SAMPLES\n"
                      << "       (2) " << PACKAGE << " MIN MAX\n"
                      << "       (3) " << PACKAGE << " SAMPLES MIN MAX\n"
                      << std::endl;
            return EXIT_SUCCESS;
          }
      }
    try
      {
        const std::string utype {"non-negative integer"};
        switch (args.size())
          {
          case 0:
            break;
          case 1:
            samples = lexical_argument_cast<unsigned>(args.at(0), utype);
            break;
          case 2:
            min_size = lexical_argument_cast<std::size_t>(args.at(1), utype);
            max_size = lexical_argument_cast<std::size_t>(args.at(2), utype);
            break;
          case 3:
            samples = lexical_argument_cast<unsigned>(args.at(0), utype);
            min_size = lexical_argument_cast<std::size_t>(args.at(1), utype);
            max_size = lexical_argument_cast<std::size_t>(args.at(2), utype);
            break;
          default:
            throw std::invalid_argument {"too many arguments"};
          }
        if (max_size < min_size)
          throw std::invalid_argument {"MAX < MIN matrix size"};
      }
    catch (const std::invalid_argument& e)
      {
        std::cerr << "error: " << e.what() << std::endl;
        return EXIT_FAILURE;
      }
    return abm.run(samples, min_size, max_size) ? EXIT_SUCCESS : EXIT_FAILURE;
  }
}
